	.586p

;	These are the selectors set up in the GDT

CODE_SEL_32		=	08h
CODE_SEL_LONG		=	10h
DATA_SEL_32		=	18h

;	Define the MSR that controls long mode

EFER_MSR		=	0c0000080h

;	Define the fixed physical address where we will put the page tables

PAGE_TABLE_BASE		=	00080000H		; PML4 goes here, 512 8 byte entries = 4k
PAGE_TABLE_PDP		=	00081000H		; PDP goes here, 512 8 byte entries
PAGE_TABLE_PD		=	00082000H		; Page directory goes here, 512 8 bytes entries

;	Define the bits in the controlling MSR

SYSCALL_ON		=	000000001h		; Enable SYSCALL instruction
LONG_MODE_ON		=	000000100h		; Enable long mode
REX32_MODE_ON		=	000000200h		; Enable Rex32

;	Define the bits in CR0

CR0_PG			=	80000000h

code	segment	use32 byte public 'code'
	assume cs:code, ds:code, es:code, ss:code

;	Enter here in 32 bit normal mode, CS L bit not set

	org	0
begin	proc	near
	jmp	near ptr Start

;	The original GDT resides up in the ROM. In order to avoid mapping that space into the paging
;	tables, we recreate it here in the lower memory area

SysGDT	label	byte
	dw	0, 0, 0, 0				; NULL gdt entry
SysRGDT	dw	0ffffh, 00000h, 09a00h, 000cfh		; 4Gb, DPL=0, L=0, D=1, Code
SysLGDT	dw	0ffffh, 00000h, 09a00h, 000Afh		; 4Gb, DPL=0, L=1, D=0, Code
DataGDT	dw	0ffffh, 00000h, 09200h, 000cfh		; 4Gb, DPL=0, Big, Data

GDTPtr	dw	4 * 8					; Limit of GDT
	dd	offset SysGDT				; Address of GDT

Start	label	near
	mov	esi, offset GDTPtr			; Point to GDTPtr variable
	db	02eh, 00fh, 001h, 016h			; lgdt cs:[esi]

	mov	ax, DATA_SEL_32
	mov	ds, ax
	mov	es, ax
	mov	fs, ax
	mov	gs, ax
	mov	ss, ax
	mov	esp, 00010000h		; Put stack at top of memory image

;	First, build up a page table set for enabling long mode. These are 2Mb page table
;		entries. CR4.PSE is assumed when long mode is active

	mov	edi, PAGE_TABLE_BASE	; Point to the PML4 base map
	mov	dword ptr [edi], PAGE_TABLE_PDP + 3	; Point to PDP, present, writable
	mov	dword ptr [edi+4], 0

	mov	edi, PAGE_TABLE_PDP	; Point to the PDP
	mov	dword ptr [edi], PAGE_TABLE_PD + 3	; Point to the page directory, present writable
	mov	dword ptr [edi+4], 0

	mov	edi, PAGE_TABLE_PD
	mov	dword ptr [edi], 0 + 00000083h		; Base 0, present, writable, 2Mb page
	mov	dword ptr [edi+4], 0

;	Now, turn LME

	xor	edx, edx
	mov	eax, LONG_MODE_ON
	mov	ecx, EFER_MSR
	wrmsr				; Write to it

;	Set up paging

	mov	eax, PAGE_TABLE_BASE
	mov	cr3, eax		; Set page base

;	Finally, turn paging on. This also turns on LMA

	mov	eax, cr0
	or	eax, CR0_PG		; Turn paging on
	mov	cr0, eax		; ... which turns on LMA

	db	0eah			; JMPF
	dd	offset LMAOn
	dw	CODE_SEL_LONG

;	Entry here is with long mode enabled and active, paging on, 1 4Mb page

LMAOn	label	near

;****	Insert any test code below this point
;
  mov rax, 0000000000000001h     ; initialize AX (64 bits)
  mov rbx, 0000000000000000h     ; initialize BX (64 bits) 
  mov ecx, eax                   ; Save off AX
  add rax, rbx                   ; Calculate next Fibonacci number
  mov ebx, ecx                   ; Move old Fib into BX
  jmp $-07h                      ; Jump back to continue algorithm

;
****	Insert any test code above this point

	xor	eax, eax		; BOP(0) is a breakpoint
	db	0f1h			; ICEBP
	hlt
begin	endp

code	ends
	end	begin

